{{- pkgname.split('.')[-1] | title | e | heading }}
.. automodule:: {{ pkgname }}

Modules in ``{{ pkgname }}`` package:

.. toctree::
   :maxdepth: 1
{% for docname in submodules %}
   {{ docname }}
{%- endfor %}
